from diffusers.optimization import TYPE_TO_SCHEDULER_FUNCTION, SchedulerType
from tango.integrations.torch.optim import LRScheduler

for k in TYPE_TO_SCHEDULER_FUNCTION.keys():
    LRScheduler.register(f"diffusers::{k.value}")(
        TYPE_TO_SCHEDULER_FUNCTION[SchedulerType(k.value)]
    )



import os
from pathlib import Path

from tango.common.testing import TangoTestCase


class TextualInversionTestCase(TangoTestCase):
    PROJECT_ROOT = (Path(__file__).parent / ".." / ".." / "..").resolve()
    """
    Root of the git repository.
    """

    # to run test suite with finished package, which does not contain
    # tests & fixtures, we must be able to look them up somewhere else
    PROJECT_ROOT_FALLBACK = (
        # users wanting to run test suite for installed package
        Path(os.environ["TEXTUAL_INVERSION_SRC_DIR"])
        if "TEXTUAL_INVERSION_SRC_DIR" in os.environ
        else (
            # fallback for conda packaging
            Path(os.environ["SRC_DIR"])
            if "CONDA_BUILD" in os.environ
            # stay in-tree
            else PROJECT_ROOT
        )
    )

    MODULE_ROOT = PROJECT_ROOT_FALLBACK / "textual_inversion"
    """
    Root of the tango module.
    """

    TESTS_ROOT = PROJECT_ROOT_FALLBACK / "tests"
    """
    Root of the tests directory.
    """

    FIXTURES_ROOT = PROJECT_ROOT_FALLBACK / "test_fixtures"
    """
    Root of the test fixtures directory.
    """



from textual_inversion.transforms.image import TransformImage  # NOQA
from textual_inversion.transforms.prompt import TransformPrompt  # NOQA
from textual_inversion.transforms.examples import TransformExamples  # NOQA



import datasets
from tango import Step

from textual_inversion.transforms import TransformExamples


@Step.register("textual_inversion::transform_data")
class TransformData(Step):
    DETERMINISTIC: bool = True
    CACHEABLE: bool = True

    def run(  # type: ignore
        self,
        dataset: datasets.DatasetDict,
        example_transformer: TransformExamples,
    ) -> datasets.DatasetDict:
        dataset = dataset.with_transform(example_transformer)
        return dataset



import logging
import os
from typing import Optional

import torch
from diffusers import StableDiffusionPipeline
from tango import Step
from tango.format import Format
from tango_ext.integrations.diffusers import DiffusersPipelineFormat
from transformers.models.clip import CLIPTextModel

from textual_inversion.models import StableDiffusionModel

logger = logging.getLogger(__name__)


def save_progress(
    text_encoder: CLIPTextModel,
    placeholder_token: str,
    placeholder_token_id: int,
    output_dir: str,
) -> None:
    if not os.path.exists(output_dir):
        logger.info(f"Make directory to {output_dir}")
        os.makedirs(output_dir)

    learned_embeds = text_encoder.get_input_embeddings().weight[placeholder_token_id]  # type: ignore
    learned_embeds_dict = {placeholder_token: learned_embeds.detach().cpu()}

    learned_embeds_path = os.path.join(output_dir, "learned_embeds.bin")
    logger.info(f"Saving embeddings to {learned_embeds_path}")
    torch.save(learned_embeds_dict, learned_embeds_path)


@Step.register("textual_inversion::create_pipeline")
class CreatePipeline(Step):
    DETERMINISTIC: bool = True
    CACHEABLE: Optional[bool] = False

    FORMAT: Format = DiffusersPipelineFormat()

    def run(  # type: ignore
        self,
        model_name: str,
        model: StableDiffusionModel,
        output_dir: str,
        placeholder_token: str,
    ) -> StableDiffusionPipeline:
        assert isinstance(model, StableDiffusionModel)

        pipeline = StableDiffusionPipeline.from_pretrained(
            model_name,
            text_encoder=model.text_encoder,
            vae=model.vae,
            unet=model.unet,
            tokenizer=model.tokenizer,
        )

        # Also save the newly trained embeddings
        save_progress(
            text_encoder=model.text_encoder,
            placeholder_token=placeholder_token,
            placeholder_token_id=model.placeholder_token_id,
            output_dir=output_dir,
        )
        return pipeline



from typing import Any, List, TypedDict

import torch
from PIL.Image import Image as PilImage
from tango.common import Registrable

from textual_inversion.transforms import TransformImage, TransformPrompt


class RawExamples(TypedDict):
    image: List[PilImage]


class PreprocessedExamples(TypedDict):
    pixel_values: List[torch.Tensor]
    input_ids: List[torch.Tensor]


class TransformExamples(Registrable):
    def __init__(
        self,
        image_transformer: TransformImage,
        prompt_transformer: TransformPrompt,
    ) -> None:
        super().__init__()
        self.image_transformer = image_transformer
        self.prompt_transformer = prompt_transformer

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        raise NotImplementedError


@TransformExamples.register("textual_inversion")
class TextualInversionTransformExamples(TransformExamples):
    def __init__(
        self, image_transformer: TransformImage, prompt_transformer: TransformPrompt
    ) -> None:
        super().__init__(image_transformer, prompt_transformer)

    def __call__(self, examples: RawExamples) -> PreprocessedExamples:
        pixel_values = [
            self.image_transformer(image_pl=image) for image in examples["image"]
        ]
        input_ids = [self.prompt_transformer() for _ in range(len(examples["image"]))]

        assert len(pixel_values) == len(input_ids)

        return {
            "pixel_values": pixel_values,
            "input_ids": input_ids,
        }



from typing import Optional, Union

from tango.common import Lazy
from tango.integrations.torch import TorchTrainingEngine, TrainingEngine
from tango.integrations.torch.model import Model
from tango.integrations.torch.optim import LRScheduler, Optimizer
from tango.integrations.torch.train_config import TrainConfig
from textual_inversion.models import StableDiffusionModel

import torch


@TrainingEngine.register("textual_inversion")
class TextualInversionEngine(TorchTrainingEngine):
    def __init__(
        self,
        train_config: TrainConfig,
        model: Union[Model, Lazy[Model]],
        optimizer: Lazy[Optimizer],
        *,
        lr_scheduler: Optional[Lazy[LRScheduler]] = None,
        amp: bool = False,
        max_grad_norm: Optional[float] = None,
        amp_use_bfloat16: Optional[bool] = None
    ) -> None:
        super().__init__(
            train_config,
            model,
            optimizer,
            lr_scheduler=lr_scheduler,
            amp=amp,
            max_grad_norm=max_grad_norm,
            amp_use_bfloat16=amp_use_bfloat16,
        )
        assert isinstance(self.model, StableDiffusionModel), type(self.model)

    def _construct_optimizer(self, optimizer: Lazy[Optimizer]) -> Optimizer:
        # Initialize the optimizer
        optim: Optimizer = optimizer.construct(
            params=self.model.text_encoder.get_input_embeddings().parameters()  # type: ignore
        )
        return optim

    def backward(self, loss: torch.Tensor) -> None:
        super().backward(loss)

        grads = self.model.text_encoder.get_input_embeddings().weight.grad  # type: ignore
        index_grads_to_zero = (
            torch.arange(len(self.model.tokenizer)) != self.model.placeholder_token_id  # type: ignore
        )
        grads.data[index_grads_to_zero, :] = grads.data[index_grads_to_zero, :].fill_(0)



import random
from typing import Literal, Union, get_args

import torch
from tango.common import Registrable
from transformers import CLIPTokenizer, CLIPTokenizerFast

from textual_inversion.templates import (
    IMAGENET_STYLE_TEMPLATES_SMALL,
    IMAGENET_TEMPLATES_SMALL,
)

LearnableProperty = Literal["object", "style"]


class TransformPrompt(Registrable):
    def __init__(
        self,
        placeholder_string: str,
        tokenizer: Union[CLIPTokenizer, CLIPTokenizerFast],
        learnable_property: LearnableProperty,
    ) -> None:
        super().__init__()
        assert isinstance(tokenizer, (CLIPTokenizer, CLIPTokenizerFast))

        self.placeholder_string = placeholder_string
        self.tokenizer = tokenizer

        assert learnable_property in get_args(LearnableProperty), learnable_property
        self.templates = (
            IMAGENET_STYLE_TEMPLATES_SMALL
            if learnable_property == "style"
            else IMAGENET_TEMPLATES_SMALL
        )

    def __call__(self) -> torch.Tensor:
        raise NotImplementedError


@TransformPrompt.register("textual_inversion")
class TextualInversionTransformPrompt(TransformPrompt):
    def __init__(
        self,
        placeholder_string: str,
        tokenizer: Union[CLIPTokenizer, CLIPTokenizerFast],
        learnable_property: LearnableProperty = "object",
    ) -> None:
        super().__init__(placeholder_string, tokenizer, learnable_property)

    def __call__(self) -> torch.Tensor:
        text = random.choice(self.templates).format(self.placeholder_string)

        input_ids = self.tokenizer(
            text,
            padding="max_length",
            truncation=True,
            max_length=self.tokenizer.model_max_length,
            return_tensors="pt",
        ).input_ids[0]

        return input_ids



from typing import Any

import numpy as np
import torch
import torchvision.transforms as transforms
from diffusers.utils import PIL_INTERPOLATION
from PIL import Image
from PIL.Image import Image as PilImage
from tango.common import Registrable


class TransformImage(Registrable):
    def __init__(
        self,
        image_size: int = 512,
        interpolation: str = "bicubic",
        is_center_crop: bool = False,
        flip_proba: float = 0.5,
    ) -> None:
        super().__init__()
        self.image_size = image_size
        self.interpolation = interpolation
        self.flip_proba = flip_proba
        self.is_center_crop = is_center_crop

    def __call__(self, image_pl: PilImage) -> Any:
        raise NotImplementedError


@TransformImage.register("textual_inversion")
class TextualInversionTransformImage(TransformImage):
    def __init__(
        self,
        image_size: int = 512,
        interpolation: str = "bicubic",
        is_center_crop: bool = False,
        flip_proba: float = 0.5,
    ) -> None:
        super().__init__(image_size, interpolation, is_center_crop, flip_proba)

    def __call__(self, image_pl: PilImage) -> torch.Tensor:
        if not image_pl.mode == "RGB":
            image_pl = image_pl.convert("RGB")

        image_np = np.array(image_pl).astype(np.uint8)
        if self.is_center_crop:
            crop = min(image_np.shape[0], image_np.shape[1])
            h, w = image_np.shape
            image_np = image_np[
                (h - crop) // 2 : (h + crop) // 2, (w - crop) // 2 : (w + crop) // 2
            ]

        image_pl = Image.fromarray(image_np)
        image_pl = image_pl.resize(
            (self.image_size, self.image_size),
            resample=PIL_INTERPOLATION[self.interpolation],
        )
        image_pl = transforms.RandomHorizontalFlip(p=self.flip_proba)(image_pl)

        image_np = np.array(image_pl).astype(np.uint8)
        image_np = (image_np / 127.5 - 1.0).astype(np.float32)
        image_th = torch.from_numpy(image_np).permute(2, 0, 1)
        return image_th



from typing import Final, List

IMAGENET_TEMPLATES_SMALL: Final[List[str]] = [
    "a photo of a {}",
    "a rendering of a {}",
    "a cropped photo of the {}",
    "the photo of a {}",
    "a photo of a clean {}",
    "a photo of a dirty {}",
    "a dark photo of the {}",
    "a photo of my {}",
    "a photo of the cool {}",
    "a close-up photo of a {}",
    "a bright photo of the {}",
    "a cropped photo of a {}",
    "a photo of the {}",
    "a good photo of the {}",
    "a photo of one {}",
    "a close-up photo of the {}",
    "a rendition of the {}",
    "a photo of the clean {}",
    "a rendition of a {}",
    "a photo of a nice {}",
    "a good photo of a {}",
    "a photo of the nice {}",
    "a photo of the small {}",
    "a photo of the weird {}",
    "a photo of the large {}",
    "a photo of a cool {}",
    "a photo of a small {}",
]
IMAGENET_STYLE_TEMPLATES_SMALL: Final[List[str]] = [
    "a painting in the style of {}",
    "a rendering in the style of {}",
    "a cropped painting in the style of {}",
    "the painting in the style of {}",
    "a clean painting in the style of {}",
    "a dirty painting in the style of {}",
    "a dark painting in the style of {}",
    "a picture in the style of {}",
    "a cool painting in the style of {}",
    "a close-up painting in the style of {}",
    "a bright painting in the style of {}",
    "a cropped painting in the style of {}",
    "a good painting in the style of {}",
    "a close-up painting in the style of {}",
    "a rendition in the style of {}",
    "a nice painting in the style of {}",
    "a small painting in the style of {}",
    "a weird painting in the style of {}",
    "a large painting in the style of {}",
]



from typing import Optional

from tango import Step
from tango.integrations.transformers import Tokenizer


@Step.register("textual_inversion::setup_tokenizer")
class SetupTokenizer(Step):
    DETERMINISTIC: bool = True
    CACHEABLE: Optional[bool] = True

    def run(  # type: ignore
        self,
        tokenizer: Tokenizer,
        placeholder_token: str,
    ) -> Tokenizer:
        num_added_tokens = tokenizer.add_tokens(placeholder_token)
        if num_added_tokens == 0:
            raise ValueError(
                f"The tokenizer already contains the token {placeholder_token}. "
                "Please pass a different `placeholder_token` that is not already in the tokenizer."
            )
        return tokenizer



from textual_inversion.steps.setup_tokenizer_step import SetupTokenizer  # NOQA
from textual_inversion.steps.create_pipeline import CreatePipeline  # NOQA
from textual_inversion.steps.generate_images import GenerateImages  # NOQA
from textual_inversion.steps.transform_data import TransformData  # NOQA



from typing import Optional

import torch
from diffusers import StableDiffusionPipeline
from PIL import Image
from tango import Step
from tango.integrations.torch.util import resolve_device, set_seed_all


@Step.register("textual_inversion::generate_images")
class GenerateImages(Step):
    DETERMINISTIC: bool = False
    CACHEABLE: Optional[bool] = False

    def run(  # type: ignore
        self,
        pipe: StableDiffusionPipeline,
        prompt: str,
        seed: int,
        generated_image_path: str,
        image_size: int = 512,
        grid_rows: int = 1,
        grid_cols: int = 4,
    ) -> None:
        num_images_per_prompt = grid_rows * grid_cols

        set_seed_all(seed)
        device = resolve_device()

        pipe = pipe.to(device)
        generator = torch.Generator().manual_seed(seed)

        images = pipe(
            prompt=prompt,
            width=image_size,
            height=image_size,
            generator=generator,
            num_images_per_prompt=num_images_per_prompt,
        ).images

        grid = Image.new("RGB", size=(grid_cols * image_size, grid_rows * image_size))

        for i, img in enumerate(images):
            box = (i % grid_cols * image_size, i // grid_cols * image_size)
            grid.paste(img, box=box)

        grid.save(generated_image_path)



import itertools
from typing import Dict, Iterable

import torch
import torch.nn as nn
import torch.nn.functional as F
from diffusers import AutoencoderKL, DDPMScheduler, UNet2DConditionModel
from tango.integrations.torch.model import Model
from transformers import CLIPTokenizer
from transformers.models.clip import CLIPTextModel


def freeze_params(params: Iterable[nn.Parameter]):
    for param in params:
        param.requires_grad = False


@Model.register("textual_inversion::stable_diffusion")
class StableDiffusionModel(Model):
    def __init__(
        self,
        model_name: str,
        tokenizer: CLIPTokenizer,
        placeholder_token: str,
        initializer_token: str,
        scale_factor: float = 0.18215,
    ) -> None:
        super().__init__()

        self.scale_factor = scale_factor
        self.tokenizer = tokenizer

        self.initializer_token_id = self.get_initializer_token_id(
            initializer_token=initializer_token,
        )
        self.placeholder_token_id = self.get_placeholder_token_id(
            placeholder_token=placeholder_token,
        )

        # load pretrained models
        self.text_encoder = self._load_text_encoder(model_name=model_name)
        self.vae = self._load_vae(model_name=model_name)
        self.unet = self._load_unet(model_name=model_name)

        # setup noise scheduler
        self.noise_scheduler = self._load_noise_scheduler(model_name=model_name)

        # setup the models
        self.resize_token_embeddings()
        self.initialize_placeholder_token()

        self.freeze_params()

    def _load_text_encoder(self, model_name: str) -> CLIPTextModel:
        return CLIPTextModel.from_pretrained(model_name, subfolder="text_encoder")  # type: ignore

    def _load_vae(self, model_name: str) -> AutoencoderKL:
        return AutoencoderKL.from_pretrained(model_name, subfolder="vae")  # type: ignore

    def _load_unet(self, model_name: str) -> UNet2DConditionModel:
        return UNet2DConditionModel.from_pretrained(model_name, subfolder="unet")  # type: ignore

    def _load_noise_scheduler(self, model_name: str) -> DDPMScheduler:
        return DDPMScheduler.from_pretrained(model_name, subfolder="scheduler")  # type: ignore

    def get_initializer_token_id(self, initializer_token: str) -> int:
        # Convert the initializer_token, placeholder_token to ids
        token_ids = self.tokenizer.encode(initializer_token, add_special_tokens=False)

        # Check if initializer_token is a single token or a sequence of tokens
        if len(token_ids) > 1:
            raise ValueError("The initializer token must be a single token.")

        return token_ids[0]

    def get_placeholder_token_id(self, placeholder_token: str) -> int:
        return self.tokenizer.convert_tokens_to_ids(placeholder_token)  # type: ignore

    def resize_token_embeddings(self) -> None:
        # Resize the token embeddings as we are adding new special tokens to the tokenizer
        self.text_encoder.resize_token_embeddings(len(self.tokenizer))

    def initialize_placeholder_token(self) -> None:
        # Initialise the newly added placeholder token with the embeddings of the initializer token
        token_embeds = self.text_encoder.get_input_embeddings().weight.data
        token_embeds[self.placeholder_token_id] = token_embeds[
            self.initializer_token_id
        ]

    def freeze_params(self) -> None:
        # Freeze vae and unet
        freeze_params(self.vae.parameters())
        freeze_params(self.unet.parameters())

        # Freeze all parameters except for the token embeddings in text encoder
        params_to_freeze = itertools.chain(
            self.text_encoder.text_model.encoder.parameters(),
            self.text_encoder.text_model.final_layer_norm.parameters(),
            self.text_encoder.text_model.embeddings.position_embedding.parameters(),
        )
        freeze_params(params_to_freeze)

    def train(self, mode: bool = True) -> "StableDiffusionModel":
        self.text_encoder.train(mode=mode)

        # Keep vae and unet in eval model as we don't train these
        self.vae.eval()
        self.unet.eval()
        return self

    def forward(
        self, pixel_values: torch.Tensor, input_ids: torch.Tensor
    ) -> Dict[str, torch.Tensor]:
        # Convert image to lantent space
        latents = self.vae.encode(pixel_values).latent_dist.sample().detach()
        latents = latents * self.scale_factor

        # Sample noise that we'll add to the latents
        noise = torch.randn(latents.shape).to(latents.device)
        bsz = latents.shape[0]
        # Sample a random timestep for each image
        timesteps = torch.randint(
            0,
            self.noise_scheduler.config.num_train_timesteps,
            (bsz,),
            device=latents.device,
        ).long()

        # Add noise to the latents according to the noise magnitude at each timestep
        # (this is the forward diffusion process)
        noisy_latents = self.noise_scheduler.add_noise(latents, noise, timesteps)

        # Get the text embedding for conditioning
        encoder_hidden_states = self.text_encoder(input_ids)[0]

        # Predict the noise residual
        noise_pred = self.unet(noisy_latents, timesteps, encoder_hidden_states).sample

        loss = F.mse_loss(noise_pred, noise, reduction="none").mean([1, 2, 3]).mean()

        return {"loss": loss}



from typing import Any, Dict, List

import torch
from tango.integrations.torch import DataCollator


@DataCollator.register("textual_inversion::custom_collator")
class CustomCollator(DataCollator[Dict[str, Any]]):
    def __call__(self, batch: List[Dict[str, Any]]) -> Dict[str, Any]:
        return {
            "pixel_values": torch.cat(
                [item["pixel_values"].unsqueeze(dim=0) for item in batch], dim=0
            ),
            "input_ids": torch.cat(
                [item["input_ids"].unsqueeze(dim=0) for item in batch], dim=0
            ),
        }
